# $Id: functions.pl,v 2.22.2.8 2010/02/12 11:53:31 mtakwale Exp $ #
# $Copyrights: Copyright (c) 2011 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $

use strict;
use warnings;
use Fcntl qw(:flock);
use POSIX;

use ag_i18n_inc;

# OUTPUT CONVENTION : 0 == true, non-0 == false

my ($mount, $umount) = ("mount", "umount");
my ($echo, $grep, $kill, $rpm, $cut) = ("echo", "grep", "kill", "rpm","cut");
my ($fuser, $fsck, $fsck_resierfs) = ("fuser", "fsck", "fsck.reiserfs");
my (@supported_fs) = ("vxfs", "ext2", "ext3", "nfs", "nfs4", "reiserfs", "bind");
my (@unsupported_mp) = ("/", "/boot", "/usr", "/proc", "/dev/pts");
my ($lock_file) = "$VCSLOG/lock/.mount_lock";
my $vcs_lock = "VCS";
my $fsadm = "/opt/VRTS/bin/fsadm";

# validate arguments
#
# input params :
#   1. file system
#   2. block device
#   3. mount point
#   4. flag to turn on logging if mount point is not present
#
# output params :
#   1. 0, if everything is OK
#      1, if file system not supported
#      2, if block device not specified
#      3, if block device is not valid
#      4, if mount point not specified
#      6, if mount point is not valid
sub validate_args
{
    my ($fs, $bd, $mp, $should_log_msg) = @_;
    my ($invalid_mp);

    my ($myrc) = is_supported_fs ($fs);
    if ($myrc) # invalid file system
    {
        VCSAG_LOG_MSG ("E",
            "File system type $fs is not supported.", 5501, "$fs");

        return 1;
    }

    unless (length ($bd)) # block device not specified
    {
        VCSAG_LOG_MSG ("E",
            "No block device name specified.", 5502);

        return 2;
    }

    if ($fs !~ /nfs/) # invalid block device
    {
    	my ($tmp_bd, $rc) = chop_slash ($bd);
	if ($rc eq 0) {
		VCSAG_LOG_MSG ("E",
				"Block device $bd is invalid since it has '/' at the end.",
				5503, "$bd");

		return 3;
	}
    }

    unless (length ($mp)) # mount point not specified
    {
        VCSAG_LOG_MSG ("E", "No mount point specified.", 5504);
        return 4;
    }


    foreach $invalid_mp (@unsupported_mp)
    {
        if ($invalid_mp eq $mp)
        {
            VCSAG_LOG_MSG ("E", 
                "Mount point $mp is not supported.", 5506, "$mp");

            return 6;
        }
    }

    return 0; # everything is OK
}


# validate fsck option
#
# input params :
#   1. fsck option
#
# output params :
#   1. 1 if fsck is invalid, else 0
sub validate_fsckopt
{
    my ($fsck_opt) = @_;

    if ($fsck_opt !~ m/-y/ && $fsck_opt !~ m/-n/)
    {
        VCSAG_LOG_MSG ("E", 
            "For FsckOpt, provide options containing either -y or -n.",
            5533);

        return 1;
    }

    return 0;
}

#validate MountOpt option
#
# input params :
#   1. MountOpt Option
#   2. Filesystem
#   3. VxFSMountLock
#
# output params :
#   1. 1 if MountOpt is invalid, else 0
sub validate_mountopt
{
    my ($mount_option, $fstype,$vxfsmntlock) = @_;
    my ($rc, $lock_val);

    # Add mount lock option also to MountOpt attribute if applicable.
    $rc = is_mntlock_supported($fstype);
    
    # Check whether the MountOpt already has a Mount Lock option.
    $lock_val = check_mount_lock($mount_option);

    if ( $rc eq 0 )
    {
        # If the MountOpt already has an option for Mount Lock,
        # If the Key is different than "VCS". Then return with
        # Error.
        if ($lock_val ne "") 
        {
            if ($vxfsmntlock eq 0)
            {
               VCSAG_LOG_MSG ("E","Mount Lock is disabled. Either set the attribute VxFSMountLock to 1 or remove mntlock from MountOpt attribute.", 5585);
               return 1;
            }
            if ($lock_val ne "VCS")
            {
               VCSAG_LOG_MSG ("E","MountOpt Attribute has a different Mount Lock Key <$lock_val>. Please change the Mount Lock Key before onlining..",5583,"$lock_val"); 
               return 1;
            }
        }

    }
    elsif (($lock_val ne "") && ($fstype eq "vxfs"))
    {
        #Will reach here if mntlock is being used on an unsupported vxfs version.
        #We dont bother if other filesystems have mntlock option or not.
        #Since we dont validate their MountOption anyway.

        VCSAG_LOG_MSG ("E", "Mount Lock is not supported for this version of vxfs. Remove mntlock option from MountOpt Attribute.",5584);
        return 1;
    }
    return 0;
}
        

# return the name of the lockfile
#
# input params : none
#
# output params :
#   1. name of the file
sub lock_file
{
    return $lock_file;
}

# lock the file passed to it and return the handle
#
# input params :
#   1. the name of the lock handle
#
# output params :
#   1. return value of the fcntl call
sub acquire_lock
{
    my ($handle) = shift;
    my ($arg, $ret_val);

    $arg = pack ("SSLLLLLLLL", F_WRLCK(), 0, 0, 0, 0, 0, 0, 0, 0, 0);
    $ret_val = fcntl($handle, F_SETLKW(), $arg);
    
    return $ret_val;
}


# unlock the passed handle
#
# input params :
#   1. lock handle to be released
#
# output params : null
sub release_lock
{
    my ($handle) = shift;
    close ($handle);
}


# unlock the passed handle and exit
#
# input params :
#   1. lock handle to be released
#   2. exit status
#
# output params : null
sub release_lock_and_exit
{
    my ($handle, $exit_status) = @_;
    close ($handle);
    exit ($exit_status);
}


# remove '/' if present at the end of the string supplied to it.
#
# input params :
#   1. string
#
# output params :
#   1. slash chopped string
#   2. 0 if chopped else 1
sub chop_slash
{
    my ($argument) = @_;
    my ($len, @arg);

    $len = length ($argument);

    @arg = split //, $argument;

    return ($argument, 1) unless ($arg[$len-1] eq '/');

    chop ($argument);
    return ($argument, 0);
}


# get raw equivalent of a device
#
# input params :
#   1. device name
#   2. file system
#
# output params :
#   1. raw equivalent of the device
sub get_raw_device
{
    my ($device, $fs) = @_;

    $device =~ s/dsk/rdsk/ if ($fs eq "vxfs");
    return $device;
}


# check if block device and mount point is mounted or not.
#
# input params :
#   1. block device
#   2. mount point
#   3. file system
#   4. flag to turn on logging
#
# ouput params :
#   1. 0, if mounted
#      1, if not mounted
#      2, if block device mounted on a different mount point
#      3, if some other block device mounted on mount point
#      4, if block device is mounted on mount point but file system is differnet
sub is_mounted
{
    my ($block_device, $mount_point, $file_system, $should_log_msg) = @_;
    my (@mount_entry, $fs, $bd, $mp);

    open (MOUNT, "$mount | $grep \"^$block_device \" |");
    while (<MOUNT>)
    {
        @mount_entry = split;
        close MOUNT if ($file_system !~ /nfs/);

        # mount-point & block-device pair matched
        $mp = $mount_entry[2];
        $fs = $mount_entry[4];
        if ($mp eq "$mount_point")
        {
            if ($fs ne "$file_system") # fs does not match
            {
                VCSAG_LOG_MSG ("E", "Block device $block_device is mounted on mount point $mount_point but with a different file system $fs.", 5508, "$block_device", "$mount_point", "$fs") if ($should_log_msg);
                return 4;
            }
            
            return 0; # mounted properly
        }

        if ( $fs !~ /nfs/)
        {
            VCSAG_LOG_MSG ("E", "Block device $block_device is mounted on a different mount point $mp.", 5509, "$block_device", "$mp") if ($should_log_msg);
            return 2; # block device mounted on different mount point
        }
    }
    close MOUNT;

    open (MOUNT, "$mount | $grep \" $mount_point \" |");
    while (<MOUNT>)
    {
        @mount_entry = split;
        close MOUNT;

        # already did a check for pair match hence not doing again
        $bd = $mount_entry[0];
        VCSAG_LOG_MSG ("E", "Mount point $mount_point is busy with some other block device $bd.", 5510, "$mount_point", "$bd") if ($should_log_msg);
        return 3; # some other device is mountes on mount point
    }
    close MOUNT;

    return 1; # not mounted
}


# mount the block device at mount point with supplied options
#
# input params :
#   1. options
#   2. block device
#   3. mount point
#
# output params : null
sub mount_fs
{
    my ($options, $block_device, $mount_point) = @_;
    VCSAG_SYSTEM ("$mount $options $block_device $mount_point");
}

# umount the mountpoints which are used as blockdevices (loopback filesystems called as bindfs)
#
# input params :
#   1. mount point
#   2. Block Device
#
# output params : null
sub umount_remounts
{
    my ($mount_point,$block_device) = @_;
    my (@mntpt_list, @mount_entry, $rc,@bd_list);
    my ($curr_mp, $curr_bd, $mntpt, $tmp_mntpt,$tmp_curr_bd);
    push (@mntpt_list, $mount_point);
    push (@bd_list, $block_device);
    open (MOUNT, "$mount |");
    while (<MOUNT>)
    {
        @mount_entry = split;
        $curr_bd = $mount_entry[0];
        $curr_mp = $mount_entry[2];
        $tmp_curr_bd = $curr_bd . "/";
        foreach $mntpt(@mntpt_list)
        {
            $tmp_mntpt = $mntpt . "/";
            if ($tmp_curr_bd =~ /^$tmp_mntpt/)
            {
                push(@mntpt_list, $curr_mp);
                push(@bd_list, $curr_bd);
            }
        }
    }

    close (MOUNT);
    while ($curr_mp = pop(@mntpt_list))
    {
        $curr_bd = pop(@bd_list);
        next if ($curr_mp eq $mount_point);
        VCSAG_LOG_MSG ("I", "Unmounting <$curr_bd> mounted on <$curr_mp>",5566,"$curr_bd","$curr_mp");
        $rc = system("$umount $curr_mp");
        next if ($rc eq 0);

        system("$fuser -15 -k -m $curr_mp");
        sleep(2);
        $rc = system("$umount $curr_mp");
        VCSAG_LOG_MSG ("I", "Unmounted  <$curr_bd> after killing processes on  <$curr_mp> using SIGTERM",5567,"$curr_bd","$curr_mp") if ($rc eq 0);
        next if ($rc eq 0);

        system("$fuser -9 -k -m $curr_mp");
        sleep(2);
        $rc = system("$umount $curr_mp");
        VCSAG_LOG_MSG ("I", "Unmounted  <$curr_bd> after killing processes on  <$curr_mp> using SIGKILL",5568,"$curr_bd","$curr_mp") if ($rc eq 0);
        next if ($rc eq 0);
    }
}

# unmount the block device / mount point
#
# inout params :
#   1. block device / mount point
#
# output params : null
sub umount_fs
{
    my ($mount_point, $type) = @_;
    if($type =~ /nfs/)
    {
        system("$umount -f $mount_point");
    }
    else
    {
        system ("$umount $mount_point");
    }
}


# mount the file system forcefully, killing all the activity
#
# input params :
#   1. options
#   2. block device
#   3. mount point
#
# output params : null
sub force_mount
{
    my ($options, $block_device, $mount_point) = @_;

    VCSAG_LOG_MSG ("N", "Trying force mount...", 5511);

    system ("$fuser -k $mount_point; $mount $options $block_device $mount_point");
}


# unmount the file system forcefully, killing all the avtivity
#
# input params :
#   1. mount point
#   2. signal
#
# output params :
#   1. 0 if force unmount succeeded else 1
sub force_umount
{
    my ($mount_point, $signal, $Type) = @_;
    
    if($Type =~ /nfs/)
    {
        
        VCSAG_LOG_MSG ("N",
            "Trying force umount for $mount_point...", 5532, "$mount_point");
        system("umount -f -l $mount_point");
    }
    else
    { 
        VCSAG_LOG_MSG ("N",
             "Trying force umount with signal $signal...", 5512, "$signal");

        system ("$fuser -$signal -k -m $mount_point");
        umount_fs ($mount_point, $Type);
    }

    my ($rc) = system ("$mount | $grep \" $mount_point \" > /dev/null");
    
    return 1 unless ($rc); # if mounted

    return 0;
}


# unmount snapshot and checkpoint of mountpoint
#
# input params :
#   1. block device
#   2. mount point
#   3. file system type
#   4. boolean : should umount snapshot or not
#   5. boolean : should umount checkpoint or not
#
# output params :
#   1. 0 if unmount succeeded else 1
sub umount_snap_ckpt
{
    my ($block_device, $mount_point, $file_system, $umount_snap, $umount_ckpt) = @_;
    my (@mount_entry, $feature, $whichumount, $try_umount, $rc);

    return 0 unless ($file_system eq "vxfs");

    open (MOUNT, "$mount |");
    while (<MOUNT>)
    {
        @mount_entry = split;
        next if ($mount_entry[4] ne "vxfs");

        if ($mount_entry[5] =~ /snapof=$mount_point[,\)\s]/)
        {
            $feature = "snapshot";
            $whichumount = "SnapUmount";
            $try_umount = ($umount_snap eq 0) ? 0 : 1;
        }
        elsif ($mount_entry[0] =~ /$block_device:/ && $mount_entry[5] =~ /[\(,\s]ckpt=/)
        {
            $feature = "checkpoint";
            $whichumount = "CkptUmount";
            $try_umount = ($umount_ckpt eq 0) ? 0 : 1;
        }
        else
        {
            next;
        }

        unless ($try_umount)
        {
            VCSAG_LOG_MSG ("W", "Cannot unmount $mount_point as $feature mount is present for the same and $whichumount is set to 0.", 5513, "$mount_point", "$feature", "$whichumount");
            close (MOUNT);
            return 1;
        }
        unlock_mount ($mount_entry[2],$mount_entry[0], $file_system);
        umount_fs ($mount_entry[2], $file_system);

        # check if the snapshot/checkpoint is still there
        $rc = system ("$mount | $grep \"^$mount_entry[0] \" > /dev/null");
        $rc = force_umount ($mount_entry[2], 9, $file_system) unless ($rc);

        if ($rc eq 1)
        {
            VCSAG_LOG_MSG ("W", "Could not unmount $feature forcefully. Cannot unmount $mount_point.", 5514, "$mount_point", "$feature");
            close (MOUNT);
            return 1;
        }
    }

    close (MOUNT);
    return 0;
}


# if vxfs file system, do log replay
#
# input params :
#   1. file system
#   2. raw device
#
# output params :
#   1. 0 if log replay performed else 1
sub log_replay
{
    my ($file_system, $device) = @_;
    if ($file_system eq "vxfs")
    {
        VCSAG_LOG_MSG ("N", "Performing log replay...", 5515);

        system ("$fsck -t $file_system $device"); 
        return 0;
    }
    return 1;
}


# run fsck command
#
# input params :
#   1. file system
#   2. fsck options
#   3. raw device
#
# output params : null
sub run_fsck
{
    my ($fstype, $fsck_opt, $device) = @_;

    return if ($fstype =~ /nfs/);
    return if ($fstype eq "bind");

    VCSAG_LOG_MSG ("N", "Running fsck...", 5516);

    if ($fstype eq "reiserfs")
    {
        $fsck_opt =~ s/-y/--fix-fixable/;
        $fsck_opt =~ s/-n/--check/;
        system ("$echo Yes | $fsck_resierfs $fsck_opt $device");
    }
    else
    {
        system ("$fsck -t $fstype $fsck_opt $device");
    }
}


# check if block device file exists or not
#
# input params :
#   1. block device
#   2. file system
#
# output params :
#   1. 0 if block device exists, else 1
sub is_block_device
{
    my ($block_device, $file_system) = @_;

    return 0 if ($file_system =~ /nfs/ || ($file_system eq "bind") || (-b $block_device));

    VCSAG_LOG_MSG ("E",
        "Block device $block_device does not exist",
        5517, "$block_device");
    return 1;
}

# check if the file system is the supported by the agent
#
# input params :
#   1. file system
#
# output parmas :
#   1. 0 if supported else 1
sub is_supported_fs
{
    my ($arg_fs) = @_;
    my ($tmp_fs);

    foreach $tmp_fs (@supported_fs)
    {
        return 0 if ($tmp_fs eq $arg_fs);
    }
    return 1;
}


# manipulate fsck option
#
# input params :
#   1. fsck option
#   2. file system
#
# output params :
#   1. manipulated fsck option
sub manipulate_fsckopt
{
    my ($fsck_opt, $file_system) = @_;

    # add file system type
    # if '-t' is not part of the fsck option, add '-t $fs'.
    $fsck_opt = "-t $file_system $fsck_opt " if ($fsck_opt !~ /-t($|\s)/);

    return $fsck_opt;
}


# manipulate mount option
#
# input params :
#   1. mount option
#   2. file system
#
# output params :
#   1. manipulated mount option
sub manipulate_mountopt
{
    my ($mount_opt, $file_system) = @_;

    # mount agent supports both interruptible soft and hard mounts
    if ($file_system =~ /nfs/)
    {
        if (length ($mount_opt))
        {
            $mount_opt .= ",intr" if ($mount_opt !~ /(^|\s|,)intr($|\s|,)/);
        }
        else
        {
            $mount_opt = "intr";
        }
    }

    if($file_system eq "bind")
    {
        if($mount_opt eq "")
        {
            $mount_opt = "bind";
        }
    }

    # if '-t' is not part of the mount option, then add '-t $fs'.
    if (length ($mount_opt))
    {
        if ($mount_opt !~ /-t/)
        {
            $mount_opt = "-t $file_system -o $mount_opt ";
        } 
    }
    else
    {
        $mount_opt = "-t $file_system " if ($mount_opt !~ /-t($|\s)/);
    }

    return $mount_opt;
}


## check if mount lock is supported
#
# input params:
#   1. FSType
#
# output params:
#   0 if mount lock is supported
#   1 if mount lock is not supported
#
sub is_mntlock_supported
{
    my ($fs) = @_;
    my $vxfs_ver = 0;
    my @vxfs_version;
    my @vxfs;
    return 1 if ($fs ne "vxfs");

    $vxfs_ver = `$rpm -q VRTSvxfs-common`;
    return 1 if ($vxfs_ver =~ /not installed/ );
    
    @vxfs = split(/\-/,$vxfs_ver);
    $vxfs_ver = $vxfs[2];
    @vxfs_version = split(/\./,$vxfs_ver);

    if ($vxfs_version[0] > 5 || ($vxfs_version[0] == 5 && $vxfs_version[1] >= 5) || ($vxfs_version[0] == 5 && $vxfs_version[2] >= 30)) 
    {
        return 0;
    }
    
    return 1;
}

## get lock key
#
# input params
#   1. MountPoint
#   2. BlockDevice
#
# output params
#   returns the key with which mount is locked or returns null
#
sub get_mount_lock
{
    my ($mntpt,$block_device) = @_;
    my ($key,$value,$mount_options,$option,$mp,@mnt_list);
   
    open (MOUNT, "$mount | $grep \"^$block_device \" |");
    while (<MOUNT>)
    {
        @mnt_list = split;
        # mount-point & block-device pair matched
        $mp = $mnt_list[2];

        if ( $mp eq $mntpt )
        {
            $mount_options = substr($mnt_list[5],1,-1);
            foreach $option (split(/,/,$mount_options))
            {
                ($key,$value) = split(/=/,$option);
                return $value if ($key eq "mntlock");
            }
        }
   
    }
    
    return "";
}

## check mount lock option
#
# input params
#   1. MountOptions
#
# output params
#   returns the key if mntlock is already used in the options
#   returns NULL if mntlock is not used in the options
#
sub check_mount_lock
{
        my ($mnt_input) = @_;
        # Search for the phrase "mntlock" to get the key.
        if ($mnt_input =~ m/mntlock=([A-Za-z0-9]+)/) {
                return $1;
        }
        # No Mount Lock
        return "";
}
## lock the mountpoint
#
# input params:
#   1. Block Device
#   2. Mount Point
#   3. FSType
#   4. Value of VxFSMountLock
#
# output params:
#   -1 if mount lock isnt supported
#   0 if lock succeeded
#   the appropriate exit status of lock command 
sub lock_mount
{
    my ($blockdevice,$mountpoint,$fstype, $vxfsmntlock) = @_;
    my ($rc ,$lock_val);

    $rc = is_mntlock_supported($fstype);

    if ( $rc eq 0 )
    {
        return -1 if ($vxfsmntlock eq 0); 
        # Check if the mount is already locked.
        $lock_val =  get_mount_lock($mountpoint,$blockdevice);
        if ( $lock_val ne "" )
        {
            #This is the case where mount options is used to lock the attribute
            #Monitor EP will take care of warning the user if the lock key is different
            return 0;
        }
    
        $rc = VCSAG_SYSTEM("$fsadm -o mntlock=\"$vcs_lock\" $mountpoint");
        return $rc;
    }
    return -1;
}


## unlock the mountpoint
#
# input params
#   1. Mount Point
#   2. Block Device
#
# output params
#   -1 if mount lock isnt supported
#   0 if unlock succeeded
#   1 if unlock failed
sub unlock_mount
{
    my ($mntpt,$block_device,$fstype) = @_;
    my ($mnt_lock,$rc);
    $rc = is_mntlock_supported($fstype);
    if ( $rc eq 0 )
    { 
        $mnt_lock = get_mount_lock($mntpt,$block_device);
        return 0 if ($mnt_lock eq ""); # already unlocked

        if ($mnt_lock ne $vcs_lock)
        {
            VCSAG_LOG_MSG ( "I", "The filesystem mounted on $mntpt is locked by key $mnt_lock. Unlocking it.", 4123, "$mntpt", "$mnt_lock");
        }
        $rc = VCSAG_SYSTEM("$fsadm -o \"mntunlock=$mnt_lock\" $mntpt");
        return $rc;
    }
    return -1;
}

return 1;
